#include <glib.h>
#include <unistd.h>

#include "apr.h"
#include "apr_strings.h"

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"
#include "util_script.h"
#include "ap_config.h"

#define BASH_OPEN_TAG		"<?bash"
#define BASH_OPEN_TAG_LEN	6

#define BASH_CLOSE_TAG		"?>"
#define BASH_CLOSE_TAG_LEN	2

struct bash_cfg_t
{
  gboolean enabled;
  gchar *chroot;
  gchar **command;
};

struct bash_t
{
  request_rec *r;
  struct bash_cfg_t *cfg;

  GMainLoop *loop;

  GPid pid;

  gchar *contents;
  gsize contents_size;

  GIOChannel *io_input;
  guint io_input_source;

  GIOChannel *io_output;
  guint io_output_source;

  GIOChannel *io_error;
  guint io_error_source;
};

typedef enum
{
  BASH_IO_IN,
  BASH_IO_OUT,
  BASH_IO_ERR
} bash_io_t;

module AP_MODULE_DECLARE_DATA bash_module;

static void *
bash_create_server_config (apr_pool_t * p, server_rec * s)
{
  struct bash_cfg_t *cfg = apr_pcalloc (p, sizeof (struct bash_cfg_t));

  memset (cfg, 0, sizeof (struct bash_cfg_t));
  return cfg;
}

static const char *
bash_cmd_enable (cmd_parms * cmd, void *dummy, int arg)
{
  struct bash_cfg_t *cfg =
    ap_get_module_config (cmd->server->module_config, &bash_module);

  cfg->enabled = arg ? TRUE : FALSE;

  return NULL;
}

static const char *
bash_cmd_chroot (cmd_parms * cmd, void *dummy, const char *chroot)
{
  struct bash_cfg_t *cfg =
    ap_get_module_config (cmd->server->module_config, &bash_module);

  cfg->chroot = apr_pstrdup (cmd->pool, chroot);
  return NULL;
}

static const char *
bash_cmd_command (cmd_parms * cmd, void *dummy, const char *args)
{
  struct bash_cfg_t *cfg =
    ap_get_module_config (cmd->server->module_config, &bash_module);

  gboolean q1, q2, q3;
  GString *opt;
  gunichar unichr;
  GList *ret = NULL;
  GList *list;
  gchar **array;
  gint i;

  opt = g_string_new (NULL);
  q1 = q2 = q3 = FALSE;

  while ((unichr = g_utf8_get_char (args)))
    {
      args = g_utf8_next_char (args);

      if (q3 == TRUE)
	{
	  opt = g_string_append_unichar (opt, unichr);
	  q3 = FALSE;
	}

      else if (unichr == '\\')
	{
	  opt = g_string_append_unichar (opt, unichr);
	  q3 = TRUE;
	}

      else if (unichr == '\"' && q2 == FALSE)
	q1 = !q1;

      else if (unichr == '\'' && q1 == FALSE)
	q2 = !q2;

      else if ((unichr == ' ' || unichr == '\t') && q1 == FALSE
	       && q2 == FALSE)
	{
	  if (opt->len)
	    {
	      ret = g_list_append (ret, g_string_free (opt, FALSE));
	      opt = g_string_new (NULL);
	    }
	}

      else
	opt = g_string_append_unichar (opt, unichr);
    }

  if (opt->len)
    ret = g_list_append (ret, g_string_free (opt, FALSE));
  else
    g_string_free (opt, TRUE);

  array = g_malloc (sizeof (gchar *) * (g_list_length (ret) + 1));
  for (i = 0, list = ret; list; list = list->next)
    array[i++] = list->data;

  array[i] = NULL;
  g_list_free (ret);

  cfg->command = array;
  return NULL;
}

static const command_rec bash_cmds[] = {
  AP_INIT_FLAG ("BashEnable", bash_cmd_enable, NULL, RSRC_CONF,
		"Enable the bash module"),

  AP_INIT_TAKE1 ("BashChroot", bash_cmd_chroot, NULL, RSRC_CONF,
		 "Set a chroot directory"),

  AP_INIT_TAKE1 ("BashCommand", bash_cmd_command, NULL, RSRC_CONF,
		 "Set bash command and arguments"),

  {NULL}
};

static void
bash_io_close (struct bash_t *bash, bash_io_t io)
{
  if (!bash)
    return;

  switch (io)
    {
    case BASH_IO_IN:
      if (bash->io_input_source)
	{
	  g_source_remove (bash->io_input_source);
	  bash->io_input_source = 0;
	}

      if (bash->io_input)
	{
	  g_io_channel_shutdown (bash->io_input, TRUE, NULL);
	  g_io_channel_unref (bash->io_input);
	  bash->io_input = NULL;
	}

      break;

    case BASH_IO_OUT:
      if (bash->io_output_source)
	{
	  g_source_remove (bash->io_output_source);
	  bash->io_output_source = 0;
	}

      if (bash->io_output)
	{
	  g_io_channel_shutdown (bash->io_output, TRUE, NULL);
	  g_io_channel_unref (bash->io_output);
	  bash->io_output = NULL;
	}

      break;

    case BASH_IO_ERR:
      if (bash->io_error_source)
	{
	  g_source_remove (bash->io_error_source);
	  bash->io_error_source = 0;
	}

      if (bash->io_error)
	{
	  g_io_channel_shutdown (bash->io_error, TRUE, NULL);
	  g_io_channel_unref (bash->io_error);
	  bash->io_error = NULL;
	}
      break;


    default:
      break;
    }
}

/* This function is called when there is something to write */
static gboolean
bash_io_out (GIOChannel * source, GIOCondition condition, struct bash_t *bash)
{
  GError *error = NULL;
  GIOStatus status;
  gsize ret;

  if ((status =
       g_io_channel_write_chars (source, bash->contents, bash->contents_size,
				 &ret, &error)) == G_IO_STATUS_NORMAL)
    status = g_io_channel_flush (source, &error);

  /* If there is an error, kill th process: */
  if (status == G_IO_STATUS_ERROR)
    {
      /* Close the channel and stop the input process: */
#ifdef G_OS_UNIX
      kill (bash->pid, SIGQUIT);
#endif
      bash_io_close (bash, BASH_IO_IN);
      return FALSE;
    }

  /* If it is fine: */
  if (status == G_IO_STATUS_NORMAL)
    {
      bash->contents_size -= ret;
      bash->contents += ret;

      /* Everything written? */
      if (bash->contents_size > 0)
	return TRUE;

      /* If yes, close the input channel: */
      bash_io_close (bash, BASH_IO_IN);
      return FALSE;
    }

  return TRUE;
}

/* This function is called when there is something to read */
static gboolean
bash_io_in (GIOChannel * source, GIOCondition condition, struct bash_t *bash)
{
  GError *error = NULL;
  gchar buffer[1024];
  GIOStatus status;
  gsize read;

  status =
    g_io_channel_read_chars (source, buffer, sizeof (buffer), &read, &error);

  if (status == G_IO_STATUS_ERROR)
    {
#ifdef G_OS_UNIX
      kill (bash->pid, SIGQUIT);
#endif
      /* Stop the output process: */
      bash_io_close (bash, BASH_IO_OUT);
      return FALSE;
    }

  /* Stop the output process: */
  if (status == G_IO_STATUS_EOF)
    {
      bash_io_close (bash, BASH_IO_OUT);
      return FALSE;
    }

  /* Good: */
  if (status == G_IO_STATUS_NORMAL)
    ap_rwrite (buffer, read, bash->r);

  return TRUE;
}

/* This function is called when there is something to read */
static gboolean
bash_io_in_err (GIOChannel * source, GIOCondition condition,
		struct bash_t *bash)
{
  GError *error = NULL;
  gchar buffer[1024];
  GIOStatus status;
  gsize read;

  status =
    g_io_channel_read_chars (source, buffer, sizeof (buffer), &read, &error);

  if (status == G_IO_STATUS_ERROR)
    {
#ifdef G_OS_UNIX
      kill (bash->pid, SIGQUIT);
#endif

      /* Stop the error process: */
      bash_io_close (bash, BASH_IO_ERR);
      return FALSE;
    }

  /* Stop the error process: */
  if (status == G_IO_STATUS_EOF)
    {
      bash_io_close (bash, BASH_IO_ERR);
      return FALSE;
    }

  /* Good: */
  if (status == G_IO_STATUS_NORMAL)
    ap_rwrite (buffer, read, bash->r);

  return TRUE;
}

/* This function is called when the process exits */
void
bash_exit (GPid pid, gint status, struct bash_t *bash)
{
  /* Remove everything from the output iochannel: */
  if (bash->io_output_source)
    {
      g_source_remove (bash->io_output_source);
      bash->io_output_source = 0;

      while (bash_io_in (bash->io_output, 0, bash) == TRUE);
    }

  /* Remove everything from the error iochannel: */
  if (bash->io_error_source)
    {
      g_source_remove (bash->io_error_source);
      bash->io_error_source = 0;

      while (bash_io_in_err (bash->io_error, 0, bash) == TRUE);
    }

  g_main_loop_quit (bash->loop);
}

static void
bash_pre_exec_chroot (struct bash_t *bash)
{
  if (bash->cfg->chroot && chroot (bash->cfg->chroot) != 0)
    {
      fprintf (stderr, "ModBash-Error chrooting to '%s': <b>%s</b>\n",
	       bash->cfg->chroot, strerror (errno));
      fflush (stderr);
    }
}

static void
bash_parse_query_string (apr_table_t * env, request_rec * r)
{
  const gchar *query_string;
  gchar **split;
  gint i;

  if (!(query_string = apr_table_get (env, "QUERY_STRING")))
    return;

  split = g_strsplit (query_string, "&", -1);
  for (i = 0; split && split[i]; i++)
    {
      gchar **var = g_strsplit (split[i], "=", 2);
      gint j;

      if (!var || !var[0] || !var[1])
	{
	  if (var)
	    g_strfreev (var);
	  continue;
	}

      ap_unescape_url (var[0]);

      for (j = 0; var[1][j]; j++)
	{
	  switch (var[1][j])
	    {
	    case '+':
	      var[1][j] = ' ';
	      break;
	    }
	  /* Other char? */
	}

      ap_unescape_url (var[1]);
      apr_table_setn (env, apr_pstrdup (r->pool, var[0]),
		      apr_pstrdup (r->pool, var[1]));
      g_strfreev (var);
    }

  g_strfreev (split);
}

static void
bash_pre_exec (struct bash_t *bash)
{
  bash_pre_exec_chroot (bash);
}

static int
bash_exec (gchar * contents, gsize contents_size, request_rec * r)
{
  struct bash_t bash;
  gint in, out, err;
  gchar **env;
  GPid pid;

  gchar **command;
  GError *error = NULL;

  memset (&bash, 0, sizeof (struct bash_t));

  bash.r = r;
  bash.contents = contents;
  bash.contents_size = contents_size;
  bash.cfg = ap_get_module_config (r->server->module_config, &bash_module);

  if (bash.cfg->command)
    command = bash.cfg->command;

  else
    {
      command = apr_pcalloc (r->pool, sizeof (gchar *) * 2);
      command[0] = "bash";
      command[1] = NULL;
    }

  ap_add_common_vars (r);
  ap_add_cgi_vars (r);

  bash_parse_query_string (r->subprocess_env, r);
  env = (gchar **) ap_create_environment (r->pool, r->subprocess_env);

  if (g_spawn_async_with_pipes
      (NULL, command, env, G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_SEARCH_PATH,
       (GSpawnChildSetupFunc) bash_pre_exec, &bash, &pid, &in, &out, &err,
       &error) == FALSE)
    {
      ap_log_rerror (APLOG_MARK, APLOG_ERR, APR_EEXIST, r,
		     "error executing the command '%s': %s", command[0],
		     error->message);

      g_error_free (error);
      return HTTP_INTERNAL_SERVER_ERROR;
    }

#ifdef G_OS_WIN32
  bash.io_input = g_io_channel_win32_new_fd (in);
#else
  bash.io_input = g_io_channel_unix_new (in);
#endif

  g_io_channel_set_encoding (bash.io_input, NULL, NULL);
  g_io_channel_set_flags (bash.io_input,
			  g_io_channel_get_flags (bash.
						  io_input) |
			  G_IO_FLAG_NONBLOCK, NULL);

  bash.io_input_source =
    g_io_add_watch (bash.io_input, G_IO_OUT, (GIOFunc) bash_io_out, &bash);

#ifdef G_OS_WIN32
  bash.io_output = g_io_channel_win32_new_fd (out);
  bash.io_error = g_io_channel_win32_new_fd (err);
#else
  bash.io_output = g_io_channel_unix_new (out);
  bash.io_error = g_io_channel_unix_new (err);
#endif

  g_io_channel_set_encoding (bash.io_output, NULL, NULL);
  g_io_channel_set_flags (bash.io_output,
			  g_io_channel_get_flags (bash.
						  io_output) |
			  G_IO_FLAG_NONBLOCK, NULL);

  bash.io_output_source =
    g_io_add_watch (bash.io_output, G_IO_IN, (GIOFunc) bash_io_in, &bash);

  g_io_channel_set_encoding (bash.io_error, NULL, NULL);
  g_io_channel_set_flags (bash.io_error,
			  g_io_channel_get_flags (bash.
						  io_error) |
			  G_IO_FLAG_NONBLOCK, NULL);

  bash.io_error_source =
    g_io_add_watch (bash.io_error, G_IO_IN, (GIOFunc) bash_io_in_err, &bash);

  g_child_watch_add (bash.pid, (GChildWatchFunc) bash_exit, &bash);

  /* Glib Loop: */
  bash.loop = g_main_loop_new (NULL, FALSE);
  g_main_loop_run (bash.loop);

  /* Close everything: */
  g_main_loop_unref (bash.loop);
  return 0;
}

static GString *
bash_string_append (GString * string, gchar * str)
{
  g_string_append_printf (string, "echo -n \"");

  for (; str && *str; str++)
    {
      if (*str == '\"')
	string = g_string_append_c (string, '\\');

      string = g_string_append_c (string, *str);
    }

  string = g_string_append (string, "\";");

  return string;
}

static int
bash_handler (request_rec * r)
{
  gchar *contents, **split;
  GString *string;
  gsize length;
  gint i, ret;

  GError *error = NULL;

  struct bash_cfg_t *cfg =
    ap_get_module_config (r->server->module_config, &bash_module);

  if (cfg->enabled == FALSE)
    return DECLINED;

  if (r->header_only)
    return OK;

  if (r->finfo.filetype == APR_NOFILE)
    return HTTP_NOT_FOUND;

  if (g_file_get_contents (r->filename, &contents, &length, &error) == FALSE)
    {
      ap_log_rerror (APLOG_MARK, APLOG_ERR, APR_EEXIST, r,
		     "error opening the file '%s': %s", r->filename,
		     error->message);

      g_error_free (error);
      return HTTP_FORBIDDEN;
    }

  if (!(split = g_strsplit (contents, BASH_OPEN_TAG, -1)) || !split[0]
      || !split[1])
    {
      ap_rwrite (contents, length, r);
      g_free (contents);

      if (split)
	g_strfreev (split);
      return OK;
    }

  g_free (contents);

  string = g_string_new (NULL);
  string = bash_string_append (string, split[0]);

  for (i = 1; split[i]; i++)
    {
      gchar *tmp;

      for (tmp = split[i];
	   tmp && *tmp && strncmp (tmp, BASH_CLOSE_TAG, BASH_CLOSE_TAG_LEN);
	   tmp++)
	string = g_string_append_c (string, *tmp);

      string = g_string_append_c (string, '\n');

      if (tmp && *tmp)
	{
	  tmp += BASH_CLOSE_TAG_LEN;
	  string = bash_string_append (string, tmp);
	}
    }

  g_strfreev (split);

  length = string->len;
  contents = g_string_free (string, FALSE);

  ret = bash_exec (contents, length, r);
  g_free (contents);

  return ret;
}

static void
bash_register_hooks (apr_pool_t * p)
{
  ap_hook_handler (bash_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

// API hooks
module AP_MODULE_DECLARE_DATA bash_module = {
  STANDARD20_MODULE_STUFF,
  NULL,				/* create per-dir    config structures */
  NULL,				/* merge  per-dir    config structures */
  bash_create_server_config,	/* create per-server config structures */
  NULL,				/* merge  per-server config structures */
  bash_cmds,			/* table of config file commands       */
  bash_register_hooks		/* register hooks                      */
};
